from typing import List
import logging
from math import prod
import numpy as np
from numpy.typing import NDArray
from cse587Autils.utils.check_probability import check_probability

logger = logging.getLogger(__name__)


def safe_exponentiate(base: [int, float],
                      exponent: [int, float]) -> [int, float]:
    """Safely exponentiates the base to the given exponent

    :param base: The base to exponentiate
    :type base: int, float]
    :param exponent: the power to raise the base to
    :type exponent: int, float]
    :return: The result of raising the given base to the given exponent.
        If the exponent is zero, the result is 1.
    :rtype: [int, float]
    """
    logger.debug('Exponentiating %s to the power of %s', base, exponent)
    if not isinstance(base, (int, float, np.int_, np.float_)):
        raise ValueError('The base must be an int or float')
    if not isinstance(exponent, (int, float, np.int_, np.float_)):
        raise ValueError('The exponent must be an int or float')
    if base == 0 and exponent != 0:
        raise ValueError('If the count is 0, the probability should not 0. A face '
                         'with no probability of being rolled should not '
                         'be observed. If this occurs as an initial guess, '
                         'you should revise your initial guess.')

    return 1 if exponent == 0 else base ** exponent


class Die:
    """
    A class used to represent a dice with n faces, each with probability p.

    :param face_probs: The probabilities of the faces
    :type face_probs: list of float

    :Examples:

    >>> face_probs = [1/6]*6
    >>> my_die = Die(face_probs)
    >>> len(my_die)
    6

    >>> # access the probability of a face. Remember that python is zero-indexed!
    >>> my_die[0]
    0.16666666666666666

    >>> # accessing a probability that is negative or outside of the range of the
    >>> # die will raise an IndexError
    >>> my_die[7]  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    IndexError

    >>> # rolling the die will return a random face
    >>> # remember python is 0 indexed AND the upper bound is exclusive, meaning
    >>> # in order to get the numbers 0, 1, 2, 3, 4, 5, you need to pass in the
    >>> # range 0, 6
    >>> my_die.roll() in set(range(0, 6))
    True
    """

    def __init__(self, face_probs: List[float] = None):
        """See class docstring for details"""
        self._face_probs = []
        logger.debug('constructing Dice object with '
                     'face_probs: %s', face_probs)
        if face_probs is not None:
            self.face_probs = face_probs

    @property
    def face_probs(self) -> List[float]:
        """
        The getter of the `face_probs` attribute.

        :return: The face probabilities (face_probs)
        :rtype: list of float
        """
        return self._face_probs

    @face_probs.setter
    def face_probs(self, value: List[float]):
        """
        The setter of the `face_probs` attribute.

        :param value: The new face probabilities (face_probs)
        :type value: list of float
        """
        valid_value = check_probability(value)
        logger.info('setting face_probs to %s', valid_value)
        self._face_probs = valid_value

    def __repr__(self) -> str:
        """
        The representation of the `Die` object.

        :return: The representation of the `Die` object
        :rtype: str

        :Exmaples:

        >>> face_probs = [1/6]*6
        >>> my_die = Die(face_probs)
        >>> my_die
        Die([0.1667, 0.1667, 0.1667, 0.1667, 0.1667, 0.1667])
        """
        return f'Die({[round(x,4) for x in self.face_probs]})'

    def __len__(self) -> int:
        """
        The getter of the length of the `face_probs` attribute.

        :return: The length of the face probabilities (face_probs)
        :rtype: int

        :Examples:

        >>> face_probs = [1/6]*6
        >>> my_die = Die(face_probs)
        >>> len(my_die)
        6
        """
        return len(self._face_probs)

    def __getitem__(self, index: int) -> float:
        """
        Return the probability of the face at a given index.

        :param index: The index of the face
        :type index: int
        :raise TypeError: If the index is not an integer
        :raise IndexError: If the index is out of range
        :return: The probability of the face at the given index
        :rtype: float

        :Examples:

        >>> face_probs = [1/6]*6
        >>> my_die = Die(face_probs)
        >>> my_die[0]
        0.16666666666666666
        """
        if not isinstance(index, int):
            raise TypeError("The index must be an integer.")
        if index < 0 or index >= len(self.face_probs):
            raise IndexError("The index must be between 0 and "
                             f"{len(self.face_probs) - 1}.")
        return self.face_probs[index]

    def __sub__(self, other: 'Die') -> float:
        """
        Subtract the face probabilities (face_probs) of one Die from another
            Die. Sum the result. This provides a measure of distance between
            two Die.

        :param other: The other Die
        :type other: Die
        :return: The sum of the differences between the face probabilities
            (face_probs) of two Die
        :rtype: float

        :raise TypeError: If the other Die is not a Die
        :raise ValueError: If the other Die has a different number of faces

        :Examples:

        >>> face_probs = [1/6]*6
        >>> my_die = Die(face_probs)
        >>> other_die = Die(face_probs)
        >>> my_die - other_die
        0.0
        """
        if not isinstance(other, Die):
            raise TypeError("The other Die must be a Die.")
        if len(self) != len(other):
            raise ValueError("The other Die must have the same "
                             "number of faces.")
        return sum(abs(self[i] - other[i]) for i in range(len(self)))

    def roll(self, seed: int = None) -> int:
        """
        Return the result of rolling the die.

        :param seed: The seed for the random number generator
        :type seed: int, optional

        :return: The result of rolling the die
        :rtype: int

        :Examples:

        >>> import numpy as np
        >>> np.random.seed(42)
        >>> face_probs = [1/6]*6
        >>> my_die = Die(face_probs)
        >>> my_die.roll() in set(range(0, 6))
        True
        """
        if seed:
            np.random.seed(seed)
        return np.random.choice(range(len(self)), p=self.face_probs)

    def likelihood(self, observed_data: NDArray[np.int_]) -> List[float]:
        """Calculate the likelihood of the observed data given the Die 
            face probabilities (face_probs).

        :param observed_data: A list of observed face counts where the index
            of each element corresponds to the face, and the count is
            the number of times that face was observed. The sum of the counts
            is the number of times the die was rolled.
        :type observed_data: NDArray[:py:class:`numpy.int_`]

        :return: The likelihood of the observed bin counts given the 
            face probabilities.
        :rtype: float

        :raises TypeError: If the face counts is not a 
            numpy array or a base python list.
        :raises ValueError: If the face counts is an empty list.

        :Examples:

        >>> import numpy as np
        >>> face_probs = [1/4]*4
        >>> my_die = Die(face_probs)
        >>> observed_data = np.array([1]*4)
        >>> round(my_die.likelihood(observed_data), 4)
        0.0039
        """
        # check input data types
        if not isinstance(observed_data, (np.ndarray, list)):
            raise TypeError('The face counts must be a numpy array '
                            'or a base python list.')
        # check that input data are not empty lists
        if len(observed_data) < 1:
            raise ValueError('The face counts must have at least one element.')
        # if the length of the face counts is greater than the face
        # probabilities, then only calculate the likelihood over the number
        # of faces in the probability list. Warn the user of this.
        if len(observed_data) > len(self.face_probs):
            logger.warning('The number of observed faces is greater than the '
                           'number of probabilities. The extra observed faces '
                           'will be ignored.')
            observed_data = observed_data[:len(self.face_probs)]
        # iterate over each face count and find the probability of rolling that
        # face that many times. Multiply the probabilities of observing each
        # face for the total likelihood.
        likelihood = prod(map(safe_exponentiate,
                              self.face_probs,
                              observed_data))

        return likelihood


if __name__ == "__main__":
    import doctest
    doctest.testmod()



import logging
import logging.config
import os


def configure_logging(level=logging.INFO,
                      to_file: bool = False,
                      filename: str = None):
    """Configure logging to print to the console and/or a file.

    :param level: The logging level, defaults to logging.INFO
    :type level: int, optional
    :param to_file: Whether to log to a file, defaults to False
    :type to_file: bool, optional
    :param filename: The name of the file to log to, defaults to None
    :type filename: str, optional

    :raises ValueError: If the logging level is not valid
    :raises ValueError: If to_file is not a boolean
    :raises ValueError: If filename is not a string
    :raises ValueError: If filename is not provided when to_file is True
    :raises FileNotFoundError: If the directory for the log file does not exist

    Configure Logging Examples
    --------------------------
    >>> configure_logging()
    >>> configure_logging(level=logging.DEBUG)
    >>> configure_logging(level=logging.INFO)
    """
    if level not in [logging.DEBUG,
                     logging.INFO,
                     logging.WARNING,
                     logging.ERROR,
                     logging.CRITICAL]:
        raise ValueError("The logging level must be one of logging.DEBUG, "
                         "logging.INFO, logging.WARNING, logging.ERROR, "
                         "or logging.CRITICAL.")
    
    if not isinstance(to_file, bool):
        raise ValueError("to_file must be a boolean.")
    
    if to_file:
        if not filename:
            raise ValueError("filename must be provided when to_file is True.")
        if not isinstance(filename, str):
            raise ValueError("filename must be a string.")
        # the `or` operator returns the first value if it is truthy
        log_file_dir = os.path.dirname(filename) or '.'
        if not os.path.isdir(log_file_dir):
            raise FileNotFoundError("The directory for the log file does not "
                                    "exist.")

    handlers = {}

    if to_file and filename:
        handlers['file'] = {
            'class': 'logging.FileHandler',
            'filename': filename,
            'mode': 'a',
            'formatter': 'detailed',  # specify the formatter
        }
    else:
        handlers['console'] = {
            'class': 'logging.StreamHandler',
            'formatter': 'detailed',  # specify the formatter
        }

    LOGGING_CONFIG = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {  # define your formatters here
            'detailed': {
                'format': '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
            },
            'simple': {
                'format': '%(levelname)s - %(message)s'
            }
        },
        'handlers': handlers,
        'root': {
            'handlers': list(handlers.keys()),
            'level': level,
        },
    }

    logging.config.dictConfig(LOGGING_CONFIG)




import pytest
from cse587Autils.utils.flatten_2d_list import flatten_2d_list


def test_flatten_2d_list():
    assert flatten_2d_list([[1, 2], [3, 4], [5, 6]]) == [1, 2, 3, 4, 5, 6]
    assert flatten_2d_list([[1], [2], [3]]) == [1, 2, 3]
    assert flatten_2d_list([]) == []

    with pytest.raises(TypeError):
        flatten_2d_list("not a list")



"""Classes and functions for the siteEM assignment series"""
import sys
import logging
import copy
from typing import List
from cse587Autils.utils.check_probability import check_probability
from cse587Autils.utils.flatten_2d_list import flatten_2d_list
from cse587Autils.utils.euclidean_distance_lists \
    import euclidean_distance_lists

logger = logging.getLogger(__name__)


class SiteModel:
    """
    A class for storing and managing parameters for a simple probabilistic
    model of transcription factor binding sites in a genome.

    :Example:

    >>> site_prior = 0.2
    >>> site_probs = [[0.25, 0.25, 0.25, 0.25], [0.1, 0.2, 0.3, 0.4]]
    >>> background_probs = [1/4]*4
    >>> sm = SiteModel(site_prior, site_probs, background_probs)
    >>> len(sm)
    2
    """

    def __init__(self,
                 site_prior: float = None,
                 site_probs: List[List[float]] = None,
                 background_probs: List[float] = None,
                 precision: int = sys.float_info.dig,
                 tolerance: float = 1e-10) -> None:
        """
        See the property documentation for details on each parameter.

        Note that if site_prior is set, background_prior is automatically set
            to 1 - site_prior.
        """
        self._precision = precision
        self._tolerance = tolerance
        if site_prior is not None:
            self.site_prior = site_prior
        if site_probs:
            self.site_probs = site_probs
        if background_probs:
            self.background_probs = background_probs

    @property
    def precision(self) -> int:
        """
        Get or set The number of digits which can accurately represent
            floating-point numbers. This is used to round the priors. By
            default, SiteModel objects have precision set to
            sys.float_info.dig, which is the runtime machine's precision.

        :return: Precision for floating-point operations.
        :rtype: int

        :Raises:
            - TypeError: If the precision is not an int.
            - ValueError: If the precision is less than 0.

        :Example:

        >>> sm = SiteModel()
        >>> sm.precision = 15
        >>> sm.precision
        15
        """
        return self._precision

    @precision.setter
    def precision(self, precision: int):
        if not isinstance(precision, int):
            raise TypeError('The precision must be an int.')
        if precision < 0:
            raise ValueError('The precision must be greater than 0.')
        self._precision = precision

    @property
    def tolerance(self) -> float:
        """
        Get or set the tolerance for checking probabilities. Defaults to 1e-10
            if not explicitly provided in constructor.

        :return: Tolerance for checking probabilities.
        :rtype: float

        :Raises:
            - TypeError: If the tolerance is not a float.
            - ValueError: If the tolerance is less than 0 or greater than 1.

        :Example:

        >>> sm = SiteModel()
        >>> sm.tolerance = 1e-10
        >>> sm.tolerance
        1e-10
        """
        return self._tolerance

    @tolerance.setter
    def tolerance(self, tolerance: float):
        if not isinstance(tolerance, (float, int)):
            raise TypeError('The tolerance must be a float.')
        if tolerance < 0 or tolerance > 1:
            raise ValueError('The tolerance must be between 0 and 1.')
        self._tolerance = tolerance

    @property
    def site_prior(self) -> float:
        """
        Prior probability of a bound site, defaults to None if not provided in
            constructor. If site_prior is set, background_prior will be set to
            1 - site_prior. This automatic update of the opposite prior occurs
            when either site_prior or background_prior are updated in an
            instance of SiteModel, also.

        :return: Prior probability of a bound site.
        :rtype: float

        :Example:

        >>> sm = SiteModel()
        >>> sm.site_prior = 0.2
        >>> round(sm.site_prior,1)
        0.2
        >>> round(sm.background_prior,1)
        0.8
        """
        try:
            return self._site_prior
        except AttributeError:
            logger.warning('site_prior not set')
            return None

    @site_prior.setter
    def site_prior(self, prior: float):
        logger.warning('Setting site_prior will also set background_prior to '
                       '1 - site_prior')
        rounded_site_prior = round(prior, self.precision)
        rounded_background_prior = round(1.0 - prior, self.precision)
        check_probability([rounded_site_prior, rounded_background_prior],
                          tolerance=self.tolerance)
        self._site_prior = rounded_site_prior
        self._background_prior = rounded_background_prior

    @property
    def background_prior(self) -> float:
        """
        Get or set the prior probability of a non-bound site. Defaults to None
            if site_prior is not passed in the object constructor. However, if
            site_prior is passed in the constructor, or if background_prior is
            updated, site_prior will be set to 1 - background_prior.

        :return: Prior probability of a non-bound site.
        :rtype: float

        :Example:

        >>> sm = SiteModel()
        >>> sm.background_prior = 0.8
        >>> round(sm.background_prior,1)
        0.8
        >>> round(sm.site_prior,1)
        0.2
        """
        try:
            return self._background_prior
        except AttributeError:
            logger.warning('background_prior not set')
            return None

    @background_prior.setter
    def background_prior(self, prior: float):
        logger.warning('Setting background_prior will also set site_prior to '
                       '1 - background_prior')
        rounded_site_prior = round(1 - prior, self.precision)
        rounded_background_prior = round(prior, self.precision)
        check_probability([rounded_site_prior, rounded_background_prior],
                          tolerance=self.tolerance)
        self._site_prior = rounded_site_prior
        self._background_prior = rounded_background_prior

    @property
    def site_probs(self) -> List[List[float]]:
        """
        List of lists containing probabilities for each base in bound sites.
            Each sublist will be length 4 and represents the probability of
            observing each base (A, C, G, T) at the given position in a bound
            site. Defaults to None if not provided in constructor. The length
            of site_probs is the length of the site sequence and is provided
            by len(SiteModel). If not explicitly passed in the constructor,
            site_probs defaults to `None`.

        :return: A list of lists containing probabilities for each base in
            bound sites.
        :rtype: list[list[float]]

        :Raises:
            - TypeError: If the value is not a list of lists.
            - ValueError: If each sublist is ont length 4.

        :Example:

        >>> sm = SiteModel()
        >>> sm.site_probs = [[0.25, 0.25, 0.25, 0.25], [0.1, 0.2, 0.3, 0.4]]
        >>> sm.site_probs[1]
        [0.1, 0.2, 0.3, 0.4]
        """
        try:
            return self._site_probs
        except AttributeError:
            logger.warning('site_probs not set')
            return None

    @site_probs.setter
    def site_probs(self, site_probs: List[List[float]]):
        if not isinstance(site_probs, list):
            raise TypeError('The value must be a list of lists.')
        for site_prob in site_probs:
            if not isinstance(site_prob, list):
                raise TypeError('Each element in `site_probs` must be a list')
            if not len(site_prob) == 4:
                raise ValueError('Each element in `site_probs` must '
                                 'be length 4.')
            check_probability(site_prob, tolerance=self.tolerance)
        self._site_probs = site_probs

    @property
    def background_probs(self) -> List[float]:
        """
        List containing the background probabilities for each base. This is a
            list of length four, where each element represents the probability
            of observing each base (A, C, G, T) in the background. It is a
            simplifying assumption that the probability of observing each base
            is independent of the position in the genome. Defaults to None if
            `background_probs` is not provided in the constructor.

        :return: A list containing the background probabilities for each base.
        :rtype: list[float]

        :Example:

        >>> sm = SiteModel()
        >>> sm.background_probs = [0.25, 0.25, 0.25, 0.25]
        >>> sm.background_probs
        [0.25, 0.25, 0.25, 0.25]
        """
        try:
            return self._background_probs
        except AttributeError:
            logger.warning('background_probs not set')
            return None

    @background_probs.setter
    def background_probs(self, background_probs: List[float]):
        if not isinstance(background_probs, list):
            raise TypeError('The value must be a list.')
        if not len(background_probs) == 4:
            raise ValueError('The value must be length 4.')
        check_probability(background_probs, tolerance=self.tolerance)
        self._background_probs = background_probs

    def __repr__(self) -> str:
        """
        Generate an unambiguous string representation of the SiteModel
            instance.

        This string representation is intended for debugging and should be
            able to recreate the object if passed to the eval() function.

        :return: A string representation of the SiteModel instance.
        :rtype: str

        :Example:

        >>> site_prior = 0.2
        >>> site_probs = [[0.25, 0.25, 0.25, 0.25], [0.1, 0.2, 0.3, 0.4]]
        >>> background_probs = [1/4]*4
        >>> sm = SiteModel(site_prior, site_probs, background_probs)
        >>> repr(sm)
        'SiteModel(site_prior=0.2, site_probs=[[0.25, 0.25, 0.25, 0.25], [0.1, 0.2, 0.3, 0.4]], background_probs=[0.25, 0.25, 0.25, 0.25])'
        """
        return (f'SiteModel(site_prior={self.site_prior}, '
                f'site_probs={self.site_probs}, '
                f'background_probs={self.background_probs})')

    def __str__(self) -> str:
        """
        Generate a human-readable string representation of the SiteModel
            instance.

        This string representation is intended for end-users and provides an
        easily interpretable description of the SiteModel instance.

        :return: A human-readable string representation of the SiteModel
            instance.
        :rtype: str

        :Example:

        >>> site_prior = 0.2
        >>> site_probs = [[0.25, 0.25, 0.25, 0.25], [0.1, 0.2, 0.3, 0.4]]
        >>> background_probs = [1/4]*4
        >>> sm = SiteModel(site_prior, site_probs, background_probs)
        >>> str(sm)  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        'SiteModel with site_prior: 0.2, background_prior: 0.8, site_probs: [[0.25, 0.25, 0.25, 0.25], [0.1, 0.2, 0.3, 0.4]], background_probs: [0.25, 0.25, 0.25, 0.25]'
        """
        return (f'SiteModel with site_prior: {self.site_prior}, '
                f'background_prior: {self.background_prior}, '
                f'site_probs: {self.site_probs}, '
                f'background_probs: {self.background_probs}')

    def __len__(self) -> int:
        """Return the number of positions in the site sequence.

        :return: Number of positions in the site sequence.
        :rtype: int

        :Example:

        >>> sm = SiteModel()
        >>> sm.site_probs = [[0.25, 0.25, 0.25, 0.25], [0.1, 0.2, 0.3, 0.4]]
        >>> len(sm)
        2
        """
        try:
            return len(self.site_probs)
        except AttributeError:
            logger.warning('site_probs not set')
            return None

    def __sub__(self, other: 'SiteModel') -> float:
        """
        Calculate the absolute difference between two SiteModels.

        The difference is calculated by taking the sum of the euclidean
        distance between [site_prior, background_prior], the euclidean
        distance between each site probability, and the euclidean distance
        between each background probability.

        :param other: The other SiteModel to compare to.
        :type other: SiteModel

        :raises TypeError: If the other object is not a SiteModel.
        :raises ValueError: If the two SiteModels do not have the same length
            site_probs or if any of the attributes are not set.

        :return: The absolute difference between the two SiteModels.
        :rtype: float

        :Example:

        >>> sm1 = SiteModel(0.2, 
        ...                 [[0.1, 0.2, 0.3, 0.4], [0.3, 0.4, 0.3, 0.0]], 
        ...                 [0.25, 0.25, 0.25, 0.25])
        >>> sm2 = SiteModel(0.1, 
        ...                 [[0.1, 0.1, 0.1, 0.7], [0.2, 0.2, 0.2, 0.4]], 
        ...                 [0.25, 0.25, 0.25, 0.25])
        >>> sm1 - sm2
        0.7414213562373095
        """
        if not isinstance(other, SiteModel):
            raise TypeError(f"Unsupported type {type(other)} for subtraction")

        if not len(self) == len(other):
            raise ValueError("Both SiteModels must have the same "
                             "length site_probs")

        if (self.site_prior is None
            or other.site_prior is None
            or self.background_prior is None
            or other.background_prior is None
            or self.site_probs is None
            or other.site_probs is None
            or self.background_probs is None
                or other.background_probs is None):
            raise ValueError("Both SiteModels must have all parameters set")

        if len(self.site_probs) != len(other.site_probs):
            raise ValueError(
                "Both SiteModels must have the same length site_probs")

        prior_diff = euclidean_distance_lists(
            [self.site_prior, self.background_prior],
            [other.site_prior, other.background_prior]
        )
        # get the abolute difference between each site probability
        site_probs_diff = euclidean_distance_lists(
            flatten_2d_list(self.site_probs),
            flatten_2d_list(other.site_probs)
        )
        # get the absolute difference between each background probability
        background_probs_diff = euclidean_distance_lists(
            self.background_probs,
            other.background_probs
        )
        return (prior_diff
                + site_probs_diff
                + background_probs_diff)

    def __copy__(self):
        """Copy method for SiteModel

        :return: A deep copy of the SiteModel instance.
        :rtype: SiteModel
        """
        new_instance = SiteModel()
        new_instance._precision = self._precision
        new_instance._tolerance = self._tolerance

        # Deep copy the lists so that they are new instances
        new_instance._site_probs = copy.deepcopy(self._site_probs)
        new_instance._background_probs = copy.deepcopy(self._background_probs)

        # Copy the simple float attributes
        new_instance._site_prior = self._site_prior
        new_instance._background_prior = self._background_prior

        return new_instance



from math import prod
import numpy as np
from numpy.testing import assert_allclose
from cse587Autils.DiceObjects.Die import Die
from cse587Autils.DiceObjects.BagOfDice import BagOfDice


def test_bag_of_dice_constructor():
    die_priors = [0.5, 0.5]
    dice = [Die([1 / 6] * 6), Die([0.9, 0.1])]
    my_bag = BagOfDice(die_priors, dice)
    assert my_bag.die_priors == die_priors
    assert my_bag.dice == dice


def test_bag_of_dice_die_priors_setter():
    die_priors = [0.5, 0.5]
    my_bag = BagOfDice()
    my_bag.die_priors = die_priors
    assert my_bag.die_priors == die_priors


def test_bag_of_dice_dice_setter():
    dice = [Die([1 / 6] * 6), Die([0.9, 0.1])]
    my_bag = BagOfDice()
    my_bag.dice = dice
    assert my_bag.dice == dice


def test_bag_of_dice_repr():
    die_priors = [0.5, 0.5]
    dice = [Die([1 / 6] * 6), Die([0.9, 0.1])]
    my_bag = BagOfDice(die_priors, dice)
    assert repr(my_bag) == f"BagOfDice(die_priors={die_priors}, dice={dice})"


def test_bag_of_dice_len():
    die_priors = [0.5, 0.5]
    dice = [Die([1 / 6] * 6), Die([0.9, 0.1])]
    my_bag = BagOfDice(die_priors, dice)
    assert len(my_bag) == 2


def test_bag_of_dice_getitem():
    die_priors = [0.5, 0.5]
    dice = [Die([1 / 6] * 6), Die([0.9, 0.1])]
    my_bag = BagOfDice(die_priors, dice)
    assert my_bag[0] == (0.5, dice[0])
    assert my_bag[0][0] == 0.5
    assert my_bag[0][1] == dice[0]


def test_bag_of_dice_subtraction_diff():
    fair_die = Die([1 / 6] * 6)
    biased_die = Die([0.9, 0.1])
    my_bag = BagOfDice([0.5, 0.5], [fair_die, biased_die])
    other_bag = BagOfDice([0.4, 0.6], [fair_die, biased_die])
    actual = my_bag - other_bag
    expected = abs(0.5 - 0.4) + abs(0.5 - 0.6)
    assert actual == expected


def test_bag_of_dice_subtraction_same():
    """Bags are the same. Diff should be 0"""
    fair_die = Die([1 / 6] * 6)
    my_bag = BagOfDice([0.5, 0.5], [fair_die, fair_die])
    actual = my_bag - my_bag
    expected = 0
    assert actual == expected


def test_bag_of_dice_iteration():
    fair_die = Die([1 / 6] * 6)
    biased_die = Die([0.9, 0.1])
    my_bag = BagOfDice([0.5, 0.5], [fair_die, biased_die])
    for i, die_info in enumerate(my_bag):
        assert die_info == (0.5, my_bag.dice[i])


def test_bag_of_dice_draw():
    seed = 42
    fair_die = Die([1 / 6] * 6)
    biased_die = Die([0.9, 0.1])
    my_bag = BagOfDice([0.5, 0.5], [fair_die, biased_die])
    result = my_bag.draw(5, seed)
    assert sum(result) == 5


def test_bag_of_dice_likelihood():
    fair_die = Die([1 / 6] * 6)
    biased_die = Die([0.9, 0.1])
    my_bag = BagOfDice([0.5, 0.5], [fair_die, biased_die])
    observed_data = [np.array([1, 2, 3, 4, 5, 6]), np.array([1, 2])]
    likelihood = my_bag.likelihood(observed_data)
    expected_likelihood = 1.0684028233559576e-24
    assert_allclose(likelihood, expected_likelihood)


def test_likelihood_manual():
    # Set up the bag of dice with two types: fair dice and biased dice
    fair_die = Die([1/6] * 6)
    fair_die_prob = 0.4
    biased_die = Die([0.1, 0.1, 0.1, 0.1, 0.1, 0.5])
    biased_die_prob = 0.6
    bag_of_dice = BagOfDice([fair_die_prob, biased_die_prob],
                            [fair_die, biased_die])

    # Set up the observed bin counts

    sample_rolls = [np.array([0, 0, 0, 0, 0, 0]),
                       np.array([0, 0, 0, 0, 0, 1])]

    expected = 1.0
    # calculate the probabilities of each trial's bin count given the fair die
    # multiplied by the probability of drawing that die
    fair_die_probabilty_trial_1 = \
        fair_die.likelihood(sample_rolls[0]) * fair_die_prob
    fair_die_probabilty_trial_2 = \
        fair_die.likelihood(sample_rolls[1]) * fair_die_prob

    biased_die_probabilty_trial_1 = \
        biased_die.likelihood(sample_rolls[0]) * biased_die_prob
    biased_die_probabilty_trial_2 = \
        biased_die.likelihood(sample_rolls[1]) * biased_die_prob

    expected = prod([fair_die_probabilty_trial_1,
                     fair_die_probabilty_trial_2,
                     biased_die_probabilty_trial_1,
                     biased_die_probabilty_trial_2])

    # Calculate the likelihood
    actual = bag_of_dice.likelihood(sample_rolls)

    # Assert that the likelihood is 0
    assert np.isclose(actual, expected, atol=1e-10), \
        f"Expected {expected}, but got {actual}"


def test_likelihood_1():
    # Define the observed data
    observed_data = [np.zeros(6), 
                     np.zeros(6)]

    # Create the bag of dice with known probabilities
    fair_die = Die([1.0])
    bag_of_dice = BagOfDice([1.0], [fair_die])

    # Calculate the likelihood
    likelihood = bag_of_dice.likelihood(observed_data)

    # Assert that the likelihood is equal to 1
    assert likelihood == 1.0


def test_calculate_likelihood_diff_length_dice():
    fair_die = Die([1/6]*6)
    biased_die = Die([0.9, 0.1])
    bag_of_dice = BagOfDice([0.5, 0.5], [fair_die, biased_die])
    sample_rolls = [np.array([1, 2, 3, 4, 5, 6]), np.array([1, 2])]

    expected = 1
    for bin_counts in sample_rolls:
        for die_info in bag_of_dice:
            die_prior, die = die_info
            expected *= die.likelihood(bin_counts) * die_prior

    actual = bag_of_dice.likelihood(sample_rolls)

    assert np.isclose(actual, expected, atol=1e-10), \
        f"Expected {expected}, but got {actual}"


import logging
from copy import copy
import pytest
from cse587Autils.SequenceObjects.SiteModel import SiteModel


def test_valid_construction():
    site_prior = 0.2
    site_probs = [[0.25, 0.25, 0.25, 0.25], [0.1, 0.2, 0.3, 0.4]]
    background_probs = [0.25, 0.25, 0.25, 0.25]

    sm = SiteModel(site_prior, site_probs, background_probs)

    assert sm.background_prior == 0.8
    assert len(sm) == 2

    sm1 = SiteModel(1, site_probs, background_probs)
    assert sm1.background_prior == 0

    sm2 = SiteModel(0, site_probs, background_probs)
    assert sm2.background_prior == 1
    

def test_logging(caplog):
    caplog.set_level(logging.WARNING)
    sm = SiteModel()
    sm.site_prior = 0.3  # This will log a warning

    assert ("Setting site_prior will also set "
            "background_prior to 1 - site_prior" in caplog.text)


def test_invalid_precision_type():
    sm = SiteModel()
    with pytest.raises(TypeError):
        sm.precision = 'invalid_type'


def test_invalid_tolerance_type():
    sm = SiteModel()
    with pytest.raises(TypeError):
        sm.tolerance = 'invalid_type'


def test_invalid_site_probs_shape():
    sm = SiteModel()
    with pytest.raises(ValueError):
        sm.site_probs = [[0.25, 0.25, 0.25], [0.1, 0.2, 0.3, 0.4]]


def test_invalid_background_probs_length():
    sm = SiteModel()
    with pytest.raises(ValueError):
        sm.background_probs = [0.25, 0.25, 0.25]


def test_site_model_subtraction():
    sm1 = SiteModel(0.2,
                    [[0.1, 0.2, 0.5, 0.2], [0.3, 0.4, 0.2, 0.1]],
                    [0.25]*4)
    sm2 = SiteModel(0.1,
                    [[0.1, 0.1, 0.8, 0.0], [0.2, 0.2, 0.1, 0.5]],
                    [0.25]*4)
    result = sm1 - sm2
    assert result == 0.7414213562373095

def test_copy_method():
    original = SiteModel(
        site_prior=0.2,
        site_probs=[[0.25, 0.25, 0.25, 0.25], [0.1, 0.2, 0.3, 0.4]],
        background_probs=[0.25, 0.25, 0.25, 0.25]
    )

    copied = copy(original)

    # Test if the objects are different
    assert original -  copied == 0

    # Test if the attributes are the same
    assert original.site_prior == copied.site_prior
    assert original.site_probs == copied.site_probs
    assert original.background_probs == copied.background_probs
    assert original.precision == copied.precision
    assert original.tolerance == copied.tolerance

    # Test if modifications to the copy don't affect the original
    copied.site_prior = 0.3
    assert original.site_prior != copied.site_prior